/******************************************************************
Copyright 2000 by Object Craft P/L, Melbourne, Australia.

                        All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted,
provided that the above copyright notice appear in all copies and that
both that copyright notice and this permission notice appear in
supporting documentation, and that the name of Object Craft
is not be used in advertising or publicity pertaining to
distribution of the software without specific, written prior
permission.

OBJECT CRAFT DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
EVENT SHALL OBJECT CRAFT BE LIABLE FOR ANY SPECIAL, INDIRECT OR
CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF
USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
PERFORMANCE OF THIS SOFTWARE.

******************************************************************/

#include "Python.h"
#include "structmember.h"

static PyObject *error_obj;

/* ----------------------------------------------------- */

/* Declarations for objects of type Parser */

typedef enum {
    START_RECORD, START_FIELD,
    IN_FIELD, QUOTE_IN_FIELD,
    IN_QUOTED_FIELD, QUOTE_IN_QUOTED_FIELD,
    END_QUOTED_FIELD
} ParserState;

typedef struct {
    PyObject_HEAD

    ParserState state;
    PyObject *fields;		/* field list for current record */

    char *buff;			/* build current field in here */
    int buff_size;		/* size of allocated buffer */
    int buff_len;		/* length of current field */

    int ms_double_quote;	/* is " represented by ""? */
} ParserObj;

staticforward PyTypeObject ParserType;

static PyObject *raise_exception(char *msg)
{
    PyErr_SetObject(error_obj, PyString_FromString(msg));
    return NULL;
}

static void parse_save_field(ParserObj *self)
{
    PyObject *field;

    field = PyString_FromStringAndSize(self->buff, self->buff_len);
    if (field != NULL) {
	PyList_Append(self->fields, field);
	Py_XDECREF(field);
    }
    self->buff_len = 0;
}

static void parse_add_char(ParserObj *self, char c)
{
    if (self->buff_len == self->buff_size) {
	if (self->buff_size == 0) {
	    self->buff_size = 4096;
	    self->buff = malloc(self->buff_size);
	} else {
	    self->buff_size *= 2;
	    self->buff = realloc(self->buff, self->buff_size);
	}
	if (self->buff == NULL) {
	    PyErr_NoMemory();
	    return;
	}
    }
    self->buff[self->buff_len++] = c;
}

static void parse_process_char(ParserObj *self, char c)
{
    switch (self->state) {
    case START_RECORD:		/* start of record */
	if (c == '\0')
	    /* empty line - return [] */
	    break;
	/* normal character - handle as START_FIELD */
	self->state = START_FIELD;
    case START_FIELD:		/* expecting field */
	switch (c) {
	case '"':		/* start quoted field */
	    self->state = IN_QUOTED_FIELD;
	    break;
	case ',':		/* save empty field */
	    parse_save_field(self);
	    break;
	case '\0':		/* save empty field - return [fields] */
	    parse_save_field(self);
	    self->state = START_RECORD;
	    break;
	default:		/* begin new unquoted field */
	    parse_add_char(self, c);
	    self->state = IN_FIELD;
	    break;
	}
	break;

    case IN_FIELD:		/* in unquoted field */
	switch (c) {
	case ',':		/* save field - wait for new field */
	    parse_save_field(self);
	    self->state = START_FIELD;
	    break;
	case '\0':		/* end of line - return [fields] */
	    parse_save_field(self);
	    self->state = START_RECORD;
	    break;
	case '"':
	    if (self->ms_double_quote) {
		/* microsoft style double quotes; " represented by "" */
		self->state = QUOTE_IN_FIELD;
		break;
	    }
	    /* handle as normal character */
	default:		/* normal character - save in field */
	    parse_add_char(self, c);
	    break;
	}
	break;

    case QUOTE_IN_FIELD:	/* microsoft double quotes - seen a
                                   quote in an unquoted field */
	parse_add_char(self, '"');
	switch (c) {
	case '"':		/* save "" as " */
	    self->state = IN_FIELD;
	    break;
	case '\0':		/* end of line - return [fields] */
	    parse_save_field(self);
	    self->state = START_RECORD;
	    break;
	default:		/* normal character - save in field */
	    parse_add_char(self, c);
	    self->state = IN_FIELD;
	    break;
	}
	break;

    case IN_QUOTED_FIELD:	/* in quoted field */
	switch (c) {
	case '"':
	    if (self->ms_double_quote) {
		/* microsoft style double quotes; " represented by "" */
		self->state = QUOTE_IN_QUOTED_FIELD;
		break;
	    }
	    /* end of field */
	    parse_save_field(self);
	    self->state = END_QUOTED_FIELD;
	    break;
	case '\0':		/* end of line - save '\n' in field */
	    parse_add_char(self, '\n');
	    break;
	default:		/* normal character - save in field */
	    parse_add_char(self, c);
	    break;
	}
	break;

    case QUOTE_IN_QUOTED_FIELD:	/* microsoft double quotes - seen a
                                   quote in an quoted field */
	switch (c) {
	case '"':		/* save "" as " */
	    parse_add_char(self, '"');
	    self->state = IN_QUOTED_FIELD;
	    break;
	case ',':		/* save field - wait for new field */
	    parse_save_field(self);
	    self->state = START_FIELD;
	    break;
	case '\0':		/* end of line - return [fields] */
	    parse_save_field(self);
	    self->state = START_RECORD;
	    break;
	default:		/* illegal */
	    raise_exception("Comma expected after \"");
	    break;
	}
	break;

    case END_QUOTED_FIELD:	/* seen closing " on quoted field */
	switch (c) {
	case ',':		/* wait for new field */
	    self->state = START_FIELD;
	    break;
	case '\0':		/* end of line - return [fields] */
	    self->state = START_RECORD;
	    break;
	default:
	    raise_exception("Comma expected after \"");
	    break;
	}
	break;
    }
}

/* ---------------------------------------------------------------- */

static char Parser_parse__doc__[] = 
"";

static PyObject *Parser_parse(ParserObj *self, PyObject *args)
{
    char *line;

    if (!PyArg_ParseTuple(args, "s", &line))
	return NULL;

    /* Process line of text - send '\0' to processing code to
       represent end of line.  End of line which is not at end of
       string is an error. */
    while (*line) {
	char c;

	c = *line++;
	if (c == '\r') {
	    c = *line++;
	    if (c == '\0')
		/* macintosh end of line */
		break;
	    if (c == '\n') {
		c = *line++;
		if (c == '\0')
		    /* DOS end of line */
		    break;
	    }
	    return raise_exception("Newline inside string");
	}
	if (c == '\n') {
	    c = *line++;
	    if (c == '\0')
		/* unix end of line */
		break;
	    return raise_exception("Newline inside string");
	}
	parse_process_char(self, c);
	if (PyErr_Occurred())
	    return NULL;
    }
    parse_process_char(self, '\0');

    if (self->state == START_RECORD) {
	PyObject *fields = self->fields;
	self->fields = PyList_New(0);
	return fields;
    }

    Py_INCREF(Py_None);
    return Py_None;
}

/* ---------------------------------------------------------------- */

static char Parser_clear__doc__[] = 
"";

static PyObject *Parser_clear(ParserObj *self, PyObject *args)
{
    if (!PyArg_ParseTuple(args, ""))
	return NULL;

    if (self->fields)
	Py_XDECREF(self->fields);
    self->fields = PyList_New(0);
    self->buff_len = 0;
    self->state = START_RECORD;

    Py_INCREF(Py_None);
    return Py_None;
}

static struct PyMethodDef Parser_methods[] = {
    { "parse", (PyCFunction)Parser_parse, METH_VARARGS, Parser_parse__doc__ },
    { "clear", (PyCFunction)Parser_clear, METH_VARARGS, Parser_clear__doc__ },
    
    { NULL, NULL }		/* sentinel */
};

static void Parser_dealloc(ParserObj *self)
{
    if (self->buff)
	free(self->buff);
    if (self->fields)
	Py_XDECREF(self->fields);

    PyMem_DEL(self);
}

#define OFF(x) offsetof(ParserObj, x)

static struct memberlist Parser_memberlist[] = {
    { "ms_double_quote", T_INT,    OFF(ms_double_quote) },
    { "fields",          T_OBJECT, OFF(fields) },
    { NULL }  /* Sentinel */
};

static PyObject *Parser_getattr(ParserObj *self, char *name)
{
    PyObject *rv;

    rv = PyMember_Get((char *)self, Parser_memberlist, name);
    if (rv)
	return rv;
    PyErr_Clear();
    return Py_FindMethod(Parser_methods, (PyObject *)self, name);
}

static int Parser_setattr(ParserObj *self, char *name, PyObject *v)
{
    if (v == NULL) {
	PyErr_SetString(PyExc_AttributeError, "Cannot delete attribute");
	return -1;
    }
    return PyMember_Set((char *)self, Parser_memberlist, name, v);
}

static char ParserType__doc__[] = 
"";

static PyTypeObject ParserType = {
    PyObject_HEAD_INIT(&PyType_Type)
    0,				/*ob_size*/
    "Parser",			/*tp_name*/
    sizeof(ParserObj),		/*tp_basicsize*/
    0,				/*tp_itemsize*/
    /* methods */
    (destructor)Parser_dealloc,	/*tp_dealloc*/
    (printfunc)0,		/*tp_print*/
    (getattrfunc)Parser_getattr, /*tp_getattr*/
    (setattrfunc)Parser_setattr, /*tp_setattr*/
    (cmpfunc)0,			/*tp_compare*/
    (reprfunc)0,		/*tp_repr*/
    0,				/*tp_as_number*/
    0,				/*tp_as_sequence*/
    0,				/*tp_as_mapping*/
    (hashfunc)0,		/*tp_hash*/
    (ternaryfunc)0,		/*tp_call*/
    (reprfunc)0,		/*tp_str*/

    /* Space for future expansion */
    0L, 0L, 0L, 0L,
    ParserType__doc__		/* Documentation string */
};

/* End of code for Parser objects */
/* -------------------------------------------------------- */

static char csv_parser__doc__[] =
"";

static PyObject *csv_parser(PyObject *module,
			    PyObject *args, PyObject *keyword_args)
{
    static char *keywords[] = {
	"ms_double_quote", NULL
    };
    ParserObj *self = PyObject_NEW(ParserObj, &ParserType);

    if (self == NULL)
	return NULL;
    self->state = START_RECORD;
    self->fields = PyList_New(0);

    self->buff = NULL;
    self->buff_size = 0;
    self->buff_len = 0;

    self->ms_double_quote = 1;

    if (PyArg_ParseTupleAndKeywords(args, keyword_args, "|i", keywords,
				    &self->ms_double_quote))
	return (PyObject*)self;

    Py_DECREF(self);
    return NULL;
}

/* List of methods defined in the module */

static struct PyMethodDef csv_methods[] = {
    { "parser", (PyCFunction)csv_parser, METH_VARARGS | METH_KEYWORDS, csv_parser__doc__ },

    { NULL, (PyCFunction)NULL, 0, NULL } /* sentinel */
};

/* Initialization function for the module (*must* be called initcsv) */

static char csv_module_documentation[] = 
"";

void initcsv()
{
    PyObject *m, *d;

    /* Create the module and add the functions */
    m = Py_InitModule4("csv", csv_methods,
		       csv_module_documentation,
		       (PyObject*)NULL,PYTHON_API_VERSION);

    /* Add some symbolic constants to the module */
    d = PyModule_GetDict(m);

    error_obj = PyErr_NewException("csv.Error", NULL, NULL);
    PyDict_SetItemString(d, "Error", error_obj);

    /* Check for errors */
    if (PyErr_Occurred())
	Py_FatalError("can't initialize module csv");
}
